require_relative '../lib/contrackt'

describe Contrackt::Base do
  before(:all) do
    class MyContract < Contrackt::Base
      required 'foo'
    end

    class Foo < Contrackt::Base
      required 'bar'
    end

    class ComposedContract < Contrackt::Base
      required 'foo' => Foo
    end

    class ComposedWithArray < Contrackt::Base
      required 'foos' => Foo[]
    end

    class ComposedWithOptional < Contrackt::Base
      required 'foo' => Foo
      optional 'additionalStuff' => Foo
    end

    class WithCustomParser < Contrackt::Base
      required('foo').with_custom_parser { |foo| Foo.new(foo.merge('bar' => foo['bar'] + " (this was the bar)")) }
    end

    @simple_payload_hash = { "foo" => "bar" }
    @complex_payload_hash = { "foo" => { "bar" => "baz" } }
    @array_payload_hash = { "foos" => [{ "bar" => "baz" }, { "bar" => "qux" }] }
    @complex_with_optional_hash = @complex_payload_hash.merge({ "additionalStuff" => { "bar" => "and more"}})
  end

  it "can be inherited" do
    expect(true).to be true # if we get here, the inheritance(s) above didn't throw an error
  end

  context "when inherited" do
    it "should be able to be constructed with a json payload" do
      expect { MyContract.new(@simple_payload_hash) }.not_to raise_exception
    end

    it "should be able to reproduce the original json from which it was constructed" do
      contract = MyContract.new(@simple_payload_hash)
      expect(contract.to_hash).to eq @simple_payload_hash
    end

    it "should provide a getter for the key that is defined in inheritor::required" do
      contract = MyContract.new(@simple_payload_hash)
      expect(contract.foo).to eq @simple_payload_hash['foo']
    end
  end

  context "with subcontract" do
    it "should allow a hash like 'foo' => Foo in the required call" do
      expect { ComposedContract.new(@complex_payload_hash) }.not_to raise_exception
    end

    it "should properly convert its subcontract values into contract objects" do
      contract = ComposedContract.new(@complex_payload_hash)
      expect(contract.foo).to be_a Foo
      expect(contract.foo.bar).to eq 'baz'
    end

    it "should be able to reconstruct the hash from which it was constructed" do
      expect(ComposedContract.new(@complex_payload_hash).to_hash).to eq @complex_payload_hash
    end
  end

  context "with array subcontract" do
    it "should allow a hash like 'foo' => Foo[] in the required call" do
      expect { ComposedWithArray.new(@array_payload_hash) }.not_to raise_exception
    end

    it "should properly map its subcontract values into contract objects" do
      contract = ComposedWithArray.new(@array_payload_hash)
      expect(contract.foos.all? { |foo| foo.is_a? Foo }).to be true
      expect(contract.foos.map(&:bar)).to eq ['baz', 'qux']
    end

    it "should be able to reconstruct the hash from which it was constructed" do
      expect(ComposedWithArray.new(@array_payload_hash).to_hash).to eq @array_payload_hash
    end
  end

  context "with optional props" do
    context "missing in the payload" do
      it "should not throw an exception if the properties are missing" do
        expect { ComposedWithOptional.new(@complex_payload_hash) }.not_to raise_exception
      end

      it "should still allow those optional properties to be accessed, just with nil value" do
        expect(ComposedWithOptional.new(@complex_payload_hash).additionalStuff).to be nil
      end

      it "should ignore those optional properties when reconstructing the hash" do
        expect(ComposedWithOptional.new(@complex_payload_hash).to_hash).to eq @complex_payload_hash
      end
    end

    context "present in the payload" do
      it "should parse optional properties properly when they are present" do
        expect(ComposedWithOptional.new(@complex_with_optional_hash).additionalStuff).to be_a Foo
        expect(ComposedWithOptional.new(@complex_with_optional_hash).additionalStuff.bar).to eq "and more"
      end

      it "should ignore those optional properties when reconstructing the hash" do
        expect(ComposedWithOptional.new(@complex_payload_hash).to_hash).to eq @complex_payload_hash
      end
    end
  end

  context "with custom parser" do
    it "uses the custom parser to build new objects" do
      expect(WithCustomParser.new(@complex_payload_hash).foo.bar).to eq "baz (this was the bar)"
    end
  end

  context "discriminator" do
    before(:all) do
      class Card < Contrackt::Base
        required      :owner
        discriminator :type
      end

      class BusinessCard < Card
        required :phone
      end

      class CreditCard < Card
        required :account
      end
    end

    context "not a collection" do
      before(:all) do
        @discriminator_hash = { card:
          { type: "BusinessCard", owner: "Jane Doe", phone: "(555) 555-5555" }
        }

        class CardPayload < Contrackt::Base
          required card: Card
        end
      end

      it "properly applies the discriminator" do
        contract = CardPayload.new(@discriminator_hash)
        expect(contract.card.class).to be BusinessCard
        expect(contract.card.owner).to eq "Jane Doe"
      end
    end

    context "collection" do
      before(:all) do
        @discriminator_hash = {
          cards: [
            { type: "BusinessCard", owner: "Jane Doe", phone: "(555) 555-5555" },
            { type: "CreditCard", owner: "Jane Doe", account: "1234-5678-9000-0000" }
          ]
        }

        class CardsPayload < Contrackt::Base
          required cards: Card[]
        end
      end

      it "properly applies the discriminator" do
        contract = CardsPayload.new(@discriminator_hash)
        expect(contract.cards.first.class).to be BusinessCard
        expect(contract.cards.first.owner).to eq "Jane Doe"
        expect(contract.cards.first.phone).to eq "(555) 555-5555"
        expect(contract.cards.last.class).to be CreditCard
        expect(contract.cards.last.owner).to eq "Jane Doe"
        expect(contract.cards.last.account).to eq "1234-5678-9000-0000"
      end

      it "reproduces the original hash via to_hash" do
        expect(CardsPayload.new(@discriminator_hash).to_hash).to eq @discriminator_hash
      end
    end

    context "custom discriminator handling" do
      before(:all) do
        class CustomCard < Contrackt::Base
          required      :owner
          discriminator :type, 'business' => BusinessCard, 'credit' => CreditCard
        end

        class CustomCardsPayload < Contrackt::Base
          required cards: CustomCard[]
        end

        @discriminator_hash = {
          cards: [
            { type: "business", owner: "Jane Doe", phone: "(555) 555-5555" },
            { type: "credit", owner: "Jane Doe", account: "1234-5678-9000-0000" }
          ]
        }
      end
      it "properly applies the discriminator" do
        contract = CustomCardsPayload.new(@discriminator_hash)
        expect(contract.cards.first.class).to be BusinessCard
        expect(contract.cards.first.owner).to eq "Jane Doe"
        expect(contract.cards.first.phone).to eq "(555) 555-5555"
        expect(contract.cards.last.class).to be CreditCard
        expect(contract.cards.last.owner).to eq "Jane Doe"
        expect(contract.cards.last.account).to eq "1234-5678-9000-0000"
      end
    end
  end
end